from datetime import datetime, timedelta
import logging
import os
import shutil
from collections import deque
from dataclasses import dataclass
from typing import Generator, List, Mapping, Optional, Any

from alembic import migration, util
from alembic.script.revision import MultipleHeads, RevisionMap
from alembic.config import MessagingOptions
from alembic.migration import RevisionStep
from alembic.script import ScriptDirectory
from alembic.util import sqla_compat
from sqlalchemy import Table
from sqlalchemy.orm import Session
from sqlalchemy.exc import DatabaseError

logger = logging.getLogger(__name__)


@dataclass
class SqlScriptDirectoryContext:
    db: Session
    table: Table
    revision_column: Optional[str] = "version_num"


class _RevisionMap(RevisionMap):
    def get_current_head(self, branch_label: Optional[str] = None):
        try:
            return super().get_current_head(branch_label)
        except MultipleHeads as e:
            return e.heads[0]


class SqlScriptDirectory(ScriptDirectory):
    def __init__(
        self,
        dir: str,
        file_template: str = ...,
        truncate_slug_length: int | None = 40,
        version_locations: List[str] | None = None,
        sourceless: bool = False,
        output_encoding: str = "utf-8",
        timezone: str | None = None,
        hook_config: Mapping[str, str] | None = None,
        recursive_version_locations: bool = False,
        messaging_opts: MessagingOptions = ...,
        context: SqlScriptDirectoryContext | None = None,
        **kwargs,
    ) -> None:
        """
        This inits the SqlScriptDirectory

        Any leftover **kwargs get lost and is only present for compatibility reasons.
        """
        super().__init__(
            dir,
            file_template,
            truncate_slug_length,
            version_locations,
            sourceless,
            output_encoding,
            timezone,
            hook_config,
            recursive_version_locations,
            messaging_opts,
        )
        self.context = context

        self.revision_map = _RevisionMap(self._load_revisions)

        assert self.context is not None
        self.create_diff_queue()

    def get_down_revisions(self, revision: str) -> List[int]:
        connectable = self.context.db.engine

        with connectable.connect() as connection:
            with sqla_compat._ensure_scope_for_ddl(connection):
                assert connection is not None

                down_revs = connection.execute(
                    self.context.table.select()
                    .order_by(self.context.table.c[self.context.revision_column].desc())
                    .filter(
                        self.context.table.c[self.context.revision_column] > revision
                    )
                ).fetchall()

                logger.debug(f"Got {len(down_revs)} revs to downgrade to {revision}")

                return [
                    getattr(down_rev, self.context.revision_column)
                    for down_rev in down_revs
                ]

    def get_down_revisions_steps(self, steps: int) -> List[int]:
        connectable = self.context.db.engine

        with connectable.connect() as connection:
            with sqla_compat._ensure_scope_for_ddl(connection):
                assert connection is not None

                down_revs = connection.execute(
                    self.context.table.select()
                    .order_by(self.context.table.c[self.context.revision_column].desc())
                    .limit(steps)
                ).fetchall()

                logger.debug(
                    f"Got {len(down_revs)} revs to downgrade; requested {steps}"
                )

                return [
                    getattr(down_rev, self.context.revision_column)
                    for down_rev in down_revs
                ]

    def get_down_revision(self, revision: str) -> str | None:
        connectable = self.context.db.engine

        with connectable.connect() as connection:
            with sqla_compat._ensure_scope_for_ddl(connection):
                assert connection is not None

                down_rev = connection.execute(
                    self.context.table.select()
                    .filter(
                        self.context.table.c[self.context.revision_column] < revision
                    )
                    .order_by(self.context.table.c[self.context.revision_column].desc())
                ).fetchone()

                logger.debug(f"Down revision of {revision} is {down_rev}")

                if not down_rev:
                    return None

                return str(getattr(down_rev, self.context.revision_column))

    def create_diff_queue(self) -> None:
        self.missing_migrations = deque()

        connectable = self.context.db.engine

        with connectable.connect() as connection:
            with sqla_compat._ensure_scope_for_ddl(connection):
                assert connection is not None

                for revision in self._load_revisions():
                    try:
                        datetime.strptime(revision.revision, "%Y%m%d%H%M%S")
                    except ValueError:
                        logger.debug(f"The revision {revision.revision} is unsupported")
                        continue

                    try:
                        entry = connection.execute(
                            self.context.table.select().filter(
                                self.context.table.c[self.context.revision_column]
                                == revision.revision
                            )
                        ).fetchone()
                    except DatabaseError:
                        # no table exists yet
                        entry = None

                    # the revision should depend on nothing
                    revision._normalized_resolved_dependencies = ()

                    if not entry or not entry.migrated:
                        self.missing_migrations.append(revision)
                        logger.debug(
                            f"Found revision {revision.revision} not present in db!"
                        )

                self.missing_migrations = sorted(
                    self.missing_migrations, key=lambda r: int(r.revision)
                )

    def consume_missing_queue(self) -> Generator[str, None, None]:
        try:
            yield self.missing_migrations.popleft()
        except IndexError:
            raise StopIteration

    def _sync_revs(self, current_rev):
        def can_sync(rev):
            try:
                datetime.strptime(rev, "%Y%m%d%H%M%S")
            except ValueError:
                return True
            return False

        if isinstance(current_rev, tuple) and current_rev:
            current_rev, *o = current_rev

        if current_rev:
            if not can_sync(current_rev):
                raise util.CommandError("Can't sync with new migrations applied!")

        migrations = []

        for rev in super()._upgrade_revs("head", (current_rev,)):
            if can_sync(rev.revision.revision):
                migrations.append(rev)

        return migrations

    def _upgrade_revs(
        self,
        destination: str,
        current_rev: str | tuple[str],
        max_lookback_days: int,
    ) -> List[RevisionStep]:
        if isinstance(current_rev, tuple) and current_rev:
            current_rev, *o = current_rev

        if max_lookback_days != -1 and current_rev:
            current = datetime.strptime(current_rev, "%Y%m%d%H%M%S")

            oldest_allowed_revision = (
                current - timedelta(days=max_lookback_days)
            ).strftime("%Y%m%d%H%M%S")

        revisions = []

        for script in self.missing_migrations:
            if destination != "head":
                # if the scripts revision is greater than its destination it means it was created on a later date.
                if int(script.revision) > int(destination):
                    logger.debug(
                        f"Skipped revision {script.revision} as its newer than the destination {destination}"
                    )
                    continue

            if max_lookback_days != -1 and current_rev:
                # if the script revision is smaller than the oldest allowed revision skip it.
                if int(script.revision) < int(oldest_allowed_revision):
                    logger.debug(
                        f"Skipped revision {script.revision} as its older than the oldest allowed revision {oldest_allowed_revision}"
                    )
                    continue

            # if we get here no loop-guards have been trigger
            revisions.append(script)

        return [
            migration.MigrationStep.upgrade_from_script(self.revision_map, script)
            for script in revisions
        ]

    def _downgrade_revs(self, destination: str, current_rev: str) -> List[RevisionStep]:
        if destination.startswith("-"):
            steps = self.get_down_revisions_steps(int(destination[1:]))
        else:
            steps = self.get_down_revisions(destination)

        return [
            migration.MigrationStep.downgrade_from_script(
                self.revision_map, self.get_revision(str(step))
            )
            for step in steps
        ]

    def _generate_template(self, src: str, dest: str, **kw: Any) -> None:
        logger.info(f"Generating {os.path.abspath(dest)}")
        util.template_to_file(src, dest, self.output_encoding, **kw)

    def _copy_file(self, src: str, dest: str) -> None:
        logger.info(f"Generating {os.path.abspath(dest)}")
        shutil.copy(src, dest)

    def _ensure_directory(self, path: str) -> None:
        path = os.path.abspath(path)
        if not os.path.exists(path):
            logger.info(f"Creating directory {path}")
            os.makedirs(path)

    @staticmethod
    def from_parent(
        parent: ScriptDirectory, additional_options: dict
    ) -> "SqlScriptDirectory":
        return SqlScriptDirectory(**vars(parent), **additional_options)

    @staticmethod
    def from_config(config, additional_options: dict) -> "SqlScriptDirectory":
        return SqlScriptDirectory.from_parent(
            ScriptDirectory.from_config(config), additional_options
        )

    def __repr__(self):
        return f"<< {self.__class__.__name__ } >>"



from dataclasses import dataclass
from dataclasses_json import dataclass_json
from typing import List, Optional
from datetime import datetime


@dataclass_json
@dataclass
class Certificate:
    owner: Optional[str] = None
    issuer: Optional[str] = None
    serial_number: Optional[str] = None
    md5: Optional[str] = None
    sha1: Optional[str] = None
    valid_from: Optional[datetime] = None
    valid_until: Optional[datetime] = None


@dataclass_json
@dataclass
class ExtractedFile:
    name: Optional[str] = None
    file_path: Optional[str] = None
    file_size: Optional[int] = None
    sha1: Optional[str] = None
    sha256: Optional[str] = None
    md5: Optional[str] = None
    type_tags: Optional[List[str]] = None
    description: Optional[str] = None
    runtime_process: Optional[str] = None
    threat_level: Optional[int] = None
    threat_level_readable: Optional[str] = None
    av_label: Optional[str] = None
    av_matched: Optional[int] = None
    av_total: Optional[int] = None
    file_available_to_download: Optional[bool] = None


@dataclass_json
@dataclass
class FileMetadata:
    file_compositions: Optional[List[str]] = None
    imported_objects: Optional[List[str]] = None
    file_analysis: Optional[List[str]] = None
    total_file_compositions_imports: Optional[int] = None


@dataclass_json
@dataclass
class FileAccess:
    type: Optional[str] = None
    path: Optional[str] = None
    mask: Optional[str] = None


@dataclass_json
@dataclass
class CreatedFile:
    file: Optional[str] = None
    null_byte: Optional[bool] = None


@dataclass_json
@dataclass
class RegistryEntry:
    operation: Optional[str] = None
    path: Optional[str] = None
    key: Optional[str] = None
    value: Optional[str] = None
    status: Optional[str] = None
    status_human_readable: Optional[str] = None


@dataclass_json
@dataclass
class Handle:
    id: Optional[int] = None
    type: Optional[str] = None
    path: Optional[str] = None


@dataclass_json
@dataclass
class SignatureMatch:
    id: Optional[str] = None
    value: Optional[str] = None


@dataclass_json
@dataclass
class Stream:
    uid: Optional[str] = None
    file_name: Optional[str] = None
    human_keywords: Optional[str] = None
    instructions: Optional[List[str]] = None
    executed: Optional[bool] = None
    matched_signatures: Optional[List[SignatureMatch]] = None


@dataclass_json
@dataclass
class ScriptParameter:
    name: Optional[str] = None
    value: Optional[str] = None
    comment: Optional[str] = None
    argument_number: Optional[int] = None
    meaning: Optional[str] = None


@dataclass_json
@dataclass
class ScriptCall:
    cls_id: Optional[str] = None
    dispatch_id: Optional[str] = None
    status: Optional[str] = None
    result: Optional[str] = None
    parameters: Optional[List[ScriptParameter]] = None
    matched_malicious_signatures: Optional[List[str]] = None


@dataclass_json
@dataclass
class ProcessFlag:
    name: Optional[str] = None
    data: Optional[str] = None


@dataclass_json
@dataclass
class AmsiCall:
    app_name: Optional[str] = None
    filename: Optional[str] = None
    raw_script_content: Optional[str] = None


@dataclass_json
@dataclass
class Module:
    path: Optional[str] = None
    base: Optional[str] = None
    interesting: Optional[bool] = None


@dataclass_json
@dataclass
class Process:
    uid: Optional[str] = None
    parentuid: Optional[str] = None
    name: Optional[str] = None
    normalized_path: Optional[str] = None
    command_line: Optional[str] = None
    sha256: Optional[str] = None
    av_label: Optional[str] = None
    av_matched: Optional[int] = None
    av_total: Optional[int] = None
    pid: Optional[str] = None
    icon: Optional[str] = None
    file_accesses: Optional[List[FileAccess]] = None
    created_files: Optional[List[CreatedFile]] = None
    registry: Optional[List[RegistryEntry]] = None
    mutants: Optional[List[str]] = None
    handles: Optional[List[Handle]] = None
    streams: Optional[List[Stream]] = None
    script_calls: Optional[List[ScriptCall]] = None
    process_flags: Optional[List[ProcessFlag]] = None
    amsi_calls: Optional[List[AmsiCall]] = None
    modules: Optional[List[Module]] = None


@dataclass_json
@dataclass
class MitreAttackParent:
    technique: Optional[str] = None
    attck_id: Optional[str] = None
    attck_id_wiki: Optional[str] = None


@dataclass_json
@dataclass
class MitreAttack:
    tactic: Optional[str] = None
    technique: Optional[str] = None
    attck_id: Optional[str] = None
    attck_id_wiki: Optional[str] = None
    parent: Optional[MitreAttackParent] = None
    malicious_identifiers_count: Optional[int] = None
    malicious_identifiers: Optional[List[str]] = None
    suspicious_identifiers_count: Optional[int] = None
    suspicious_identifiers: Optional[List[str]] = None
    informative_identifiers_count: Optional[int] = None
    informative_identifiers: Optional[List[str]] = None


@dataclass_json
@dataclass
class Signature:
    threat_level: Optional[int] = None
    threat_level_human: Optional[str] = None
    category: Optional[str] = None
    identifier: Optional[str] = None
    type: Optional[int] = None
    relevance: Optional[int] = None
    name: Optional[str] = None
    description: Optional[str] = None
    origin: Optional[str] = None
    attck_id: Optional[str] = None
    capec_id: Optional[str] = None
    attck_id_wiki: Optional[str] = None


@dataclass_json
@dataclass
class Submission:
    submission_id: Optional[str] = None
    filename: Optional[str] = None
    url: Optional[str] = None
    created_at: Optional[datetime] = None


@dataclass_json
@dataclass
class MLModelData:
    id: Optional[str] = None
    value: Optional[str] = None


@dataclass_json
@dataclass
class MachineLearningModel:
    name: Optional[str] = None
    version: Optional[str] = None
    status: Optional[str] = None
    data: Optional[List[MLModelData]] = None
    created_at: Optional[datetime] = None


@dataclass_json
@dataclass
class ExecutableProcessMemoryAnalysis:
    filename: Optional[str] = None
    address: Optional[str] = None
    flags: Optional[str] = None
    file_process: Optional[str] = None
    file_process_pid: Optional[int] = None
    file_process_sha256: Optional[str] = None
    file_process_disc_pathway: Optional[str] = None
    verdict: Optional[str] = None


@dataclass_json
@dataclass
class AnalysisRelatedUrl:
    url: Optional[str] = None
    verdict: Optional[str] = None
    type: Optional[str] = None


@dataclass_json
@dataclass
class CrowdstrikeAI:
    executable_process_memory_analysis: Optional[
        List[ExecutableProcessMemoryAnalysis]
    ] = None
    analysis_related_urls: Optional[List[AnalysisRelatedUrl]] = None


@dataclass_json
@dataclass
class HybridAnalysisHashRecord:
    job_id: Optional[str] = None
    environment_id: Optional[int] = None
    environment_description: Optional[str] = None
    size: Optional[int] = None
    type: Optional[str] = None
    type_short: Optional[List[str]] = None
    target_url: Optional[str] = None
    state: Optional[str] = None
    error_type: Optional[str] = None
    error_origin: Optional[str] = None
    submit_name: Optional[str] = None
    md5: Optional[str] = None
    sha1: Optional[str] = None
    sha256: Optional[str] = None
    sha512: Optional[str] = None
    ssdeep: Optional[str] = None
    imphash: Optional[str] = None
    entrypoint: Optional[str] = None
    entrypoint_section: Optional[str] = None
    image_base: Optional[str] = None
    subsystem: Optional[str] = None
    image_file_characteristics: Optional[List[str]] = None
    dll_characteristics: Optional[List[str]] = None
    major_os_version: Optional[int] = None
    minor_os_version: Optional[int] = None
    av_detect: Optional[int] = None
    vx_family: Optional[str] = None
    url_analysis: Optional[bool] = None
    analysis_start_time: Optional[datetime] = None
    threat_score: Optional[int] = None
    interesting: Optional[bool] = None
    threat_level: Optional[int] = None
    verdict: Optional[str] = None
    certificates: Optional[List[Certificate]] = None
    is_certificates_valid: Optional[bool] = None
    certificates_validation_message: Optional[str] = None
    domains: Optional[List[str]] = None
    compromised_hosts: Optional[List[str]] = None
    hosts: Optional[List[str]] = None
    total_network_connections: Optional[int] = None
    total_processes: Optional[int] = None
    total_signatures: Optional[int] = None
    extracted_files: Optional[List[ExtractedFile]] = None
    file_metadata: Optional[FileMetadata] = None
    processes: Optional[List[Process]] = None
    mitre_attcks: Optional[List[MitreAttack]] = None
    network_mode: Optional[str] = None
    signatures: Optional[List[Signature]] = None
    classification_tags: Optional[List[str]] = None
    tags: Optional[List[str]] = None
    submissions: Optional[List[Submission]] = None
    machine_learning_models: Optional[List[MachineLearningModel]] = None
    crowdstrike_ai: Optional[CrowdstrikeAI] = None



from flask import Flask

from nldcsc.flask_managers.flask_app_manager import FlaskAppManager

app = Flask(__name__)


@app.route("/")
def index():
    return "SUCCESS"


class TestFlaskAppManager:
    def test_init(self):
        fam = FlaskAppManager(version="TEST_VERSION", app=app)

        assert isinstance(fam, FlaskAppManager)



from datetime import datetime
import logging
import time
from typing import Any, Tuple, Optional, Dict, Union

from alembic import util
from alembic.runtime.environment import EnvironmentContext
from alembic.runtime.migration import (
    MigrationContext,
    HeadMaintainer,
    RevisionStep,
    StampStep,
)
from alembic.util import sqla_compat
from sqlalchemy import (
    Dialect,
    Connection,
    URL,
    Engine,
    literal_column,
)
from sqlalchemy.engine import url as sqla_url
from sqlalchemy.exc import OperationalError

from nldcsc.plugins.sql_migrate.fractured_alembic.script.base import SqlScriptDirectory

from ...config.constants import schema_migrations_table

log = logging.getLogger(__name__)


class FracturedHeadMaintainer(HeadMaintainer):
    def __init__(self, context: "FracturedMigrationContext", heads: Any) -> None:
        super().__init__(context, heads)

    def stamp(self, revs):
        self.context.impl._exec(self.context._version.delete())

        for rev in revs:
            self.context.impl._exec(
                self.context._version.insert().values(
                    version_num=literal_column("'%s'" % rev)
                )
            )

    def update_migration_table(
        self, step: RevisionStep | StampStep, version: int, name: str, duration: int
    ):
        if step.is_upgrade:
            self.insert_migration(version, name, duration)
        else:
            self.remove_migration(version)

    def remove_migration(self, version: int):
        log.debug(f"Remove from migration table {version}")
        self.context.impl._exec(
            self.context._schema_migrations_table.delete().filter(
                self.context._schema_migrations_table.c.version_num == version
            )
        )

    def insert_migration(self, version: int, name: str, duration: int) -> None:
        log.debug(f"Insert into migration table {version}:{name}:{duration}")
        self.context.impl._exec(
            self.context._schema_migrations_table.insert().values(
                version_num=literal_column("'%s'" % version),
                name=literal_column("'%s'" % name),
                duration=literal_column("'%s'" % duration),
                migrated=literal_column("'%s'" % int(time.time())),
            )
        )

    def __repr__(self):
        return f"<< {self.__class__.__name__ } >>"


class FracturedMigrationContext(MigrationContext):
    """
    This class modifies the MigrationContext so that it can handle scenarios where migrations are fractured.

    A fractured migration context is the cause of working on multiple branches where changes to the db schema can happen simultaneously.
    """

    def __init__(
        self,
        dialect: Dialect,
        connection: Optional[Connection],
        opts: Dict[str, Any],
        environment_context: Optional[EnvironmentContext] = None,
    ):
        self.is_sync = opts.pop("is_sync", False)
        self.fix_head = opts.pop("fix_head", False)
        super().__init__(dialect, connection, opts, environment_context)

        self._script_directory: SqlScriptDirectory = opts.get("script_directory")
        self._schema_migrations_table = schema_migrations_table

    def _ensure_schema_migrations_table(self, purge: bool = False):
        with sqla_compat._ensure_scope_for_ddl(self.connection):
            assert self.connection is not None
            self._schema_migrations_table.create(self.connection, checkfirst=True)
            if purge:
                assert self.connection is not None
                self.connection.execute(self._schema_migrations_table.delete())

    def drop_migrations_tables(self):
        with sqla_compat._ensure_scope_for_ddl(self.connection):
            assert self.connection is not None
            try:
                self._version.drop(self.connection)
            except OperationalError:
                pass
            try:
                self._schema_migrations_table.drop(self.connection)
            except OperationalError:
                pass

    def get_full_migrations_table_content(self):
        with sqla_compat._ensure_scope_for_ddl(self.connection):
            assert self.connection is not None
            all_data = self.connection.execute(
                self._schema_migrations_table.select().order_by(
                    self._schema_migrations_table.c["id"]
                )
            ).fetchall()

        return all_data

    @classmethod
    def configure(
        cls,
        connection: Optional[Connection] = None,
        url: Optional[Union[str, URL]] = None,
        dialect_name: Optional[str] = None,
        dialect: Optional[Dialect] = None,
        environment_context: Optional[EnvironmentContext] = None,
        dialect_opts: Optional[Dict[str, str]] = None,
        opts: Optional[Any] = None,
    ) -> MigrationContext:
        if opts is None:
            opts = {}
        if dialect_opts is None:
            dialect_opts = {}

        if connection:
            if isinstance(connection, Engine):
                raise util.CommandError(
                    "'connection' argument to configure() is expected "
                    "to be a sqlalchemy.engine.Connection instance, "
                    "got %r" % connection,
                )

            dialect = connection.dialect
        elif url:
            url_obj = sqla_url.make_url(url)
            dialect = url_obj.get_dialect()(**dialect_opts)
        elif dialect_name:
            url_obj = sqla_url.make_url("%s://" % dialect_name)
            dialect = url_obj.get_dialect()(**dialect_opts)
        elif not dialect:
            raise Exception("Connection, url, or dialect_name is required.")
        assert dialect is not None
        return FracturedMigrationContext(dialect, connection, opts, environment_context)

    def run_migrations(self, **kw: Any) -> None:
        self.impl.start_migrations()

        heads: Tuple[str, ...]
        if self.purge:
            if self.as_sql:
                raise util.CommandError("Can't use --purge with --sql mode")
            self._ensure_version_table(purge=True)
            self._ensure_schema_migrations_table(purge=True)
            heads = ()
        else:
            heads = self.get_current_heads()

            dont_mutate = self.opts.get("dont_mutate", False)

            if not self.as_sql and not heads and not dont_mutate:
                self._ensure_version_table()
                self._ensure_schema_migrations_table()

        head_maintainer = FracturedHeadMaintainer(self, heads)

        assert self._migrations_fn is not None
        for step in self._migrations_fn(heads, self):
            start_time = time.time()
            with self.begin_transaction(_per_migration=True):
                if self.as_sql and not head_maintainer.heads:
                    # for offline mode, include a CREATE TABLE from the base
                    assert self.connection is not None
                    self._version.create(self.connection)
                log.info("Running %s", step)
                if self.as_sql:
                    self.impl.static_output("-- Running %s" % (step.short_log,))
                step.migration_fn(**kw)

                # previously, we wouldn't stamp per migration
                # if we were in a transaction, however given the more
                # complex model that involves any number of inserts
                # and row-targeted updates and deletes, it's simpler for now
                # just to run the operations on every version
                head_maintainer.update_to_step(step)
                for callback in self.on_version_apply_callbacks:
                    callback(
                        ctx=self,
                        step=step.info,
                        heads=set(head_maintainer.heads),
                        run_args=kw,
                    )
                duration = time.time() - start_time

                try:
                    head_maintainer.update_migration_table(
                        step,
                        int(step.insert_version_num),
                        name=step.doc,
                        duration=duration,
                    )
                except ValueError:
                    if self.is_sync:
                        pass
                    else:
                        raise util.CommandError("Invalid revision to update!")

        if self.fix_head:
            with self.begin_transaction(_per_migration=True):
                head_maintainer.stamp(self._script_directory.get_heads())

        if self.as_sql and not head_maintainer.heads:
            assert self.connection is not None
            self._version.drop(self.connection)

    # def stamp(self, script_directory: ScriptDirectory, revision: str) -> None: ...

    # NOTE this function should be updated accordingly.

    def __repr__(self):
        return f"<< {self.__class__.__name__ } >>"



from flask_sqlalchemy import SQLAlchemy

from .model_base import ModelBase


class FlaskSQLAlchemy(SQLAlchemy):
    def __init__(self, *args, **kwargs):
        base_class = kwargs.pop("model_class", None)

        if base_class is None:
            base_class = ModelBase

        kwargs["model_class"] = base_class

        super().__init__(*args, **kwargs)



from flask import Flask
from flask_sqlalchemy import SQLAlchemy

app = Flask(__name__)
app.config["SQLALCHEMY_DATABASE_URI"] = "sqlite:///:memory:"
db = SQLAlchemy(app)


@app.route("/")
def index():
    users = User.query.all()
    return [x.to_dict() for x in users]


class User(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    username = db.Column(db.String(255))
    email = db.Column(db.String(255))
    age = db.Column(db.Integer)

    def to_dict(self):
        return {"user": self.username}



from typing import List, Any, Optional

from pydantic import BaseModel


class EmailAttachment(BaseModel):
    Name: str
    Content: str
    ContentType: str = "text/plain"

    # {
    #     "Name": "readme.txt",
    #     "Content": "dGVzdCBjb250ZW50",
    #     "ContentType": "text/plain",
    # }


class Email(BaseModel):
    From: List[str] | str
    To: List[str] | str
    Subject: str
    Cc: Optional[List[str] | str] = None
    Bcc: Optional[List[str] | str] = None
    Headers: Optional[dict[str, str]] = None
    Attachments: Optional[List[EmailAttachment]] = None
    TemplateData: dict[str, Any]
    Send_at: Optional[int] = None

    # {
    #     "From": "sender@example.com",
    #     "To": "receiver@example.com",
    #     "Cc": "copied@example.com",
    #     "Bcc": "blank-copied@example.com",
    #     "Subject": "Test subject",
    #     "Headers": {"Name": "CUSTOM-HEADER", "Value": "value"},
    #     "Attachments": [
    #         {
    #             "Name": "readme.txt",
    #             "Content": "dGVzdCBjb250ZW50",
    #             "ContentType": "text/plain",
    #         },
    #         {
    #             "Name": "report.pdf",
    #             "Content": "dGVzdCBjb250ZW50",
    #             "ContentType": "application/octet-stream",
    #         },
    #     ],
    #     "TemplateData": {
    #         "date": "2025-08",
    #         "departments": [
    #             {
    #                 "title": "MT",
    #                 "content": "Lorem ipsum odor amet, consectetuer adipiscing elit.",
    #             },
    #             {
    #                 "title": "Basic Monitoring",
    #                 "content": "Lorem ipsum odor amet, consectetuer adipiscing elit. Sapien blandit "
    #                            "sagittis laoreet sociosqu congue vulputate urna vehicula.",
    #             },
    #         ],
    #     },
    #     "Send_at": 1740145440,
    # }


class BatchTemplate(Email):
    Template: str


class BatchEmail(BaseModel):
    Messages: List[BatchTemplate]

    # {
    #     "Messages": [
    #         {
    #             "Template": "weekbrief",
    #             "From": "sender@example.com",
    #             "To": "receiver@example.com",
    #             "Cc": "copied@example.com",
    #             "Bcc": "blank-copied@example.com",
    #             "Subject": "Test subject",
    #             "Headers": {"Name": "CUSTOM-HEADER", "Value": "value"},
    #             "Attachments": [
    #                 {
    #                     "Name": "readme.txt",
    #                     "Content": "dGVzdCBjb250ZW50",
    #                     "ContentType": "text/plain",
    #                 },
    #                 {
    #                     "Name": "report.pdf",
    #                     "Content": "dGVzdCBjb250ZW50",
    #                     "ContentType": "application/octet-stream",
    #                 },
    #             ],
    #             "TemplateData": {
    #                 "date": "2025-08",
    #                 "departments": [
    #                     {
    #                         "title": "MT",
    #                         "content": "Lorem ipsum odor amet, consectetuer adipiscing elit.",
    #                     },
    #                     {
    #                         "title": "Basic Monitoring",
    #                         "content": "Lorem ipsum odor amet, consectetuer adipiscing elit. Sapien blandit "
    #                         "sagittis laoreet sociosqu congue vulputate urna vehicula.",
    #                     },
    #                 ],
    #             },
    #             "Send_at": 1740145440,
    #         },
    #         {
    #             "Template": "weekbrief2",
    #             "From": "sender@example.com",
    #             "To": "receiver@example.com",
    #             "Cc": "copied@example.com",
    #             "Bcc": "blank-copied@example.com",
    #             "Subject": "Test subject",
    #             "Headers": {"Name": "CUSTOM-HEADER", "Value": "value"},
    #             "Attachments": [
    #                 {
    #                     "Name": "readme.txt",
    #                     "Content": "dGVzdCBjb250ZW50",
    #                     "ContentType": "text/plain",
    #                 },
    #                 {
    #                     "Name": "report.pdf",
    #                     "Content": "dGVzdCBjb250ZW50",
    #                     "ContentType": "application/octet-stream",
    #                 },
    #             ],
    #             "TemplateData": {
    #                 "date": "2025-08",
    #                 "departments": [
    #                     {
    #                         "title": "MT",
    #                         "content": "Lorem ipsum odor amet, consectetuer adipiscing elit.",
    #                     },
    #                     {
    #                         "title": "Basic Monitoring",
    #                         "content": "Lorem ipsum odor amet, consectetuer adipiscing elit. Sapien blandit "
    #                         "sagittis laoreet sociosqu congue vulputate urna vehicula.",
    #                     },
    #                 ],
    #             },
    #         },
    #     ]
    # }



from nldcsc.http_apis.base_class.api_base_class import ApiBaseClass


class DiscordWebhookApi(ApiBaseClass):
    DISCORD_URL = "https://discord.com"
    DISCORD_BASE_API_PATH = "api/webhooks/{webhook_id}/{token}"

    def __init__(
        self,
        webhook_id,
        token,
        proxies=None,
        user_agent="NLDCSC",
    ):
        super().__init__(
            self.DISCORD_URL,
            self.DISCORD_BASE_API_PATH.format(webhook_id=webhook_id, token=token),
            proxies,
            user_agent,
        )

    def _build_url(self, resource: str, *args, **kwargs) -> str:
        """
        Internal method to build a url to use when executing commands
        """
        out = self.baseurl

        if self.api_path:
            out += f"/{self.api_path}"
        if resource:
            out += f"/{resource}"

        return out

    def post_to_webhook(
        self,
        content: str = None,
        username: str = None,
        avatar_url: str = None,
        tts: bool = None,
        embeds: list = None,
        **kwargs,
    ):
        """
        Body could consist of (https://discord.com/developers/docs/resources/webhook#execute-webhook)
        """
        resource = ""

        if not embeds or content:
            raise TypeError(
                "Either message or embed variable should be filled; they cannot all be None"
            )

        data = {}

        if embeds:
            if not isinstance(embeds, list):
                embeds = [embeds]

            data["embeds"] = embeds

        if content:
            data["content"] = content

        if username:
            data["username"] = username

        if avatar_url:
            data["avatar_url"] = avatar_url

        if tts:
            data["tts"] = tts

        if kwargs:
            data.update(kwargs)

        return self.call("POST", resource=resource, data=data)



#!/usr/bin/env python3
import codecs
from itertools import chain
from pathlib import Path
from setuptools import find_packages, setup
from poetry.core.packages.dependency import Dependency
from poetry.factory import Factory


def parse_group_deps(pyproject_data: dict):
    """
    Get the group deps from the pyproject data.

    To get the group dep data we need to follow a specific path that may or may not exist.
    This function walks through the path and always returns a dict.

    Args:
        pyproject_data (dict): the pyproject data parsed by poetry.

    Returns:
        dict: dict containing the group dependencies
    """
    path = ["tool", "nldcsc", "group", "dependencies"]

    cursor = pyproject_data
    for p in path:
        cursor = cursor.get(p, {})

        if not isinstance(cursor, dict):
            return {}
    return cursor


def reqs(deps: list[Dependency]):
    """
    Format the poetry Dependency to the proper requirement format.

    Args:
        deps (list[Dependency]): list of dependencies.

    Returns:
        list[str]: list of properly formatted requirement strings.
    """

    return [
        f"{dep.complete_name}{'' if any(c in str(dep.constraint) for c in '>=<') else '=='}{dep.constraint}"
        for dep in deps
    ]


# -*- Long Description -*-
def long_description(readme_path: Path):
    """
    Try and read the long description from the readme file.

    Args:
        readme_path (Path): path to the readme file.

    Returns:
        str: Content of the readme file.
    """
    try:
        return codecs.open(readme_path.name, "r", "utf-8").read()
    except OSError:
        return "Long description error: Missing README.md file"


def resolve_group_deps(
    poetry_reqs: dict[str, list[str]], group_deps: dict[list[str]], group: str
) -> set[str]:
    """
    Follow the stream of group deps until a Root or multiple roots are found.

    When a root is found all requirements of that root are added to this group as dependency.

    Args:
        poetry_reqs (dict[str, list[str]]): poetry requirements data object.
        group_deps (dict[list[str]]): dependencies of all groups.
        group (str): group to resolve.

    Returns:
        set[str]: set with the requirements for the given group.
    """
    deps = set(poetry_reqs.get(group, []))

    for dep in group_deps.get(group, []):
        deps.update(resolve_group_deps(poetry_reqs, group_deps, dep))

    return deps


def parse_poetry_reqs(poetry_reqs: dict[str, list[str]], group_deps: dict[list[str]]):
    """
    Read the poetry requirements add parse them to the proper format for setup().

    Further this function creates additional extras if defined in the group deps. 'All' is always created.

    Args:
        poetry_reqs (dict[str, list[str]]): poetry requirements data object.
        group_deps (dict[list[str]]): dependencies of all groups.

    Raises:
        RecursionError: If a circular dependency exists a RecursionError is raised.

    Returns:
        tuple: contains the install_reqs (Always needed for this package), and the optional extras.
    """
    for dep in group_deps:
        try:
            poetry_reqs[dep] = list(resolve_group_deps(poetry_reqs, group_deps, dep))
        except RecursionError:
            raise RecursionError(
                f"You probably have a circular dependency for group {dep}"
            )

    intall_reqs = poetry_reqs.pop("main", [])
    poetry_reqs["all"] = list(set(chain(*poetry_reqs.values())))

    return intall_reqs, poetry_reqs


def generate_meta():
    """
    Generates all the meta info for setup()

    Returns:
        dict: dict that can be used as kwargs for setup()
    """
    poetry = Factory().create_poetry()

    install, extras = parse_poetry_reqs(
        {
            group: reqs(poetry.package.dependency_group(group).dependencies)
            for group in poetry.package.dependency_group_names()
        },
        parse_group_deps(poetry.pyproject.data),
    )

    return {
        "name": poetry.package.name,
        "packages": find_packages(exclude=["tests", "test_data"]),
        "version": str(poetry.package.version),
        "description": poetry.package.description,
        "long_description": long_description(poetry.package.readme),
        "long_description_content_type": "text/markdown",
        "author": poetry.package.author_name,
        "author_email": poetry.package.author_email,
        "url": poetry.package.repository_url,
        "license": poetry.package.license.name,
        "platforms": ["any"],
        "install_requires": install,
        "extras_require": extras,
        "python_requires": str(poetry.package.python_constraint),
        "include_package_data": True,
        "project_urls": poetry.package.custom_urls,
        "classifiers": poetry.package.classifiers,
        "entry_points": {"console_scripts": ["nldcsc = nldcsc.cli:cli"]},
    }


setup(**generate_meta())



import re
from abc import ABC, abstractmethod
from collections import namedtuple, defaultdict
from werkzeug.datastructures import CombinedMultiDict


class ServerSideDataTable(ABC):
    """
    This class holds all the logic for enabling and handling server side DataTables within the application

    """

    def __init__(self, request, backend, **kwargs):
        self.request_values: CombinedMultiDict = request.values

        self.backend = backend

        self.columns, self.ordering, self.results, self.fields, self.data_length = (
            None,
            None,
            None,
            None,
            None,
        )

        self.total, self.total_filtered, self.current_draw = 0, 0, 0

        self.filtered = {}

        self.sort = []

        self.pre_fetch_processing()

        self.post_fetch_processing()

    def output_result(self):
        return {
            "draw": int(self.current_draw),
            "recordsTotal": int(self.total),
            "recordsFiltered": int(self.total_filtered),
            "data": self.results,
        }

    def pre_fetch_processing(self):
        """
        Method to provision the necessary variables for the correct retrieval of the results from the Database
        """
        self.current_draw = int(self.request_values["draw"])

        self.data_length = self.__data_dimension()

        self.columns, self.ordering = self.__data_columns_ordering()

        for each in self.ordering:
            if self.ordering[each]["dir"] == "asc":
                self.sort.append(
                    (
                        self.columns[self.ordering[each]["column"]]["data"],
                        "asc",
                    )
                )
            else:
                self.sort.append(
                    (
                        self.columns[self.ordering[each]["column"]]["data"],
                        "desc",
                    )
                )

        self.fields = defaultdict(int)

        for each in self.columns:
            self.fields[self.columns[each]["data"]] = 1

        self.get_results_and_meta()

    @abstractmethod
    def get_results_and_meta(self):
        raise NotImplementedError

    @abstractmethod
    def fetch_results(self):
        """
        Method responsible for querying the backend and fetching the results from the database
        """
        raise NotImplementedError

    @abstractmethod
    def post_fetch_processing(self):
        raise NotImplementedError

    def __data_dimension(self):
        """
        Method responsible for retrieving the requested start and length parameters from the DataTables request values

        :return: Namedtuple 'data_length' with a start and length attribute
        :rtype: namedtuple
        """

        data_length = namedtuple("data_length", ["start", "length"])

        data_length.start = int(self.request_values["start"])
        data_length.length = int(self.request_values["length"])

        return data_length

    def __data_columns_ordering(self):
        """
        Method responsible for retrieving the column and order details from the DataTables request values

        :return: 2 Dictionaries the column details and order details
        :rtype: defaultdict, defaultdict
        """

        col = defaultdict(lambda: defaultdict(lambda: defaultdict(dict)))
        order = defaultdict(lambda: defaultdict(dict))

        col_regex = re.compile(r"columns\[(\d*)\]\[(\w*)\](?:\[(\w*)\])?")
        order_regex = re.compile(r"order\[(\d*)\]\[(\w*)\]")

        for each in sorted(self.request_values.keys()):
            check_col_match = col_regex.match(each)
            check_order_match = order_regex.match(each)
            if check_col_match:
                if check_col_match.group(2) == "search":
                    try:
                        col[check_col_match.group(1)][check_col_match.group(2)][
                            check_col_match.group(3)
                        ] = int(self.request_values[each])
                    except ValueError:
                        col[check_col_match.group(1)][check_col_match.group(2)][
                            check_col_match.group(3)
                        ] = self.request_values[each]
                else:
                    col[check_col_match.group(1)][check_col_match.group(2)] = (
                        self.request_values[each]
                    )
            if check_order_match:
                order[check_order_match.group(1)][check_order_match.group(2)] = (
                    self.request_values[each]
                )

        return col, order

    @abstractmethod
    def data_filter(self):
        """
        Method responsible for retrieving the filter values entered in the search box of the DataTables.

        :return: Prepared filter based on filterable columns and retrieved search value
        :rtype: dict
        """
        raise NotImplementedError
